@@ -1,10 +1,17 @@
 import argparse
 from dataclasses import dataclass, field
-from typing import List, Optional, Sequence, Set, Type
+import enum
+import json
+from typing import Dict, List, Optional, Sequence, Set, Tuple, Type
+from typing_extensions import TypedDict
 
 import libcst
 from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
 from libcst.codemod.visitors import AddImportsVisitor
+from libcst.metadata import CodePosition, CodeRange, PositionProvider
+
+_DEFAULT_POSITION = CodePosition(0, 0)
+_DEFAULT_CODE_RANGE = CodeRange(_DEFAULT_POSITION, _DEFAULT_POSITION)
 
 
 @dataclass
@@ -24,6 +31,16 @@ class NamedParam:
         return NamedParam(name, module, type_name)
 
 
+class PyanalyzeSuggestion(TypedDict):
+    suggested_type: str
+    imports: List[str]
+
+
+class DecoratorKind(enum.Enum):
+    asynq = 1
+    abstractmethod = 2
+
+
 @dataclass
 class State:
     annotate_optionals: List[NamedParam]
@@ -40,6 +57,10 @@ class State:
     seen_raise_statement: List[bool] = field(default_factory=lambda: [False])
     seen_yield: List[bool] = field(default_factory=lambda: [False])
     in_lambda: bool = False
+    pyanalyze_suggestions: Dict[Tuple[str, int, int], PyanalyzeSuggestion] = field(
+        default_factory=dict
+    )
+    only_without_imports: bool = False
 
 
 SIMPLE_MAGICS = {
@@ -64,10 +85,43 @@ IMPRECISE_MAGICS = {
 }
 
 
+class _SafeAction(argparse.Action):
+    def __call__(
+        self,
+        parser: argparse.ArgumentParser,
+        namespace: argparse.Namespace,
+        values: object,
+        option_string: Optional[str] = ...,
+    ) -> None:
+        namespace.none_return = True
+        namespace.scalar_return = True
+        namespace.annotate_magics = True
+
+
+class _AggressiveAction(_SafeAction):
+    def __call__(
+        self,
+        parser: argparse.ArgumentParser,
+        namespace: argparse.Namespace,
+        values: object,
+        option_string: Optional[str] = ...,
+    ) -> None:
+        super().__call__(parser, namespace, values, option_string)
+        namespace.bool_param = True
+        namespace.int_param = True
+        namespace.float_param = True
+        namespace.str_param = True
+        namespace.bytes_param = True
+        namespace.annotate_imprecise_magics = True
+
+
 class AutotypeCommand(VisitorBasedCodemodCommand):
 
     # Add a description so that future codemodders can see what this does.
     DESCRIPTION: str = "Automatically adds simple type annotations."
+    METADATA_DEPENDENCIES = (PositionProvider,)
+
+    state: State
 
     @staticmethod
     def add_args(arg_parser: argparse.ArgumentParser) -> None:
@@ -151,6 +205,30 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
                 "for __iter__)"
             ),
         )
+        arg_parser.add_argument(
+            "--pyanalyze-report",
+            type=str,
+            default=None,
+            help="Path to a pyanalyze --json-report file to use for suggested types.",
+        )
+        arg_parser.add_argument(
+            "--only-without-imports",
+            action="store_true",
+            default=False,
+            help="Only apply pyanalyze suggestions that do not require imports",
+        )
+        arg_parser.add_argument(
+            "--safe",
+            action=_SafeAction,
+            help="Apply all safe transformations",
+            nargs="?",
+        )
+        arg_parser.add_argument(
+            "--aggressive",
+            action=_AggressiveAction,
+            help="Apply all transformations that do not require arguments",
+            nargs="?",
+        )
 
     def __init__(
         self,
@@ -167,6 +245,10 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
         bytes_param: bool = False,
         float_param: bool = False,
         int_param: bool = False,
+        pyanalyze_report: Optional[str] = None,
+        only_without_imports: bool = False,
+        safe: bool = False,
+        aggressive: bool = False,
     ) -> None:
         super().__init__(context)
         param_type_pairs = [
@@ -176,6 +258,30 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
             (int_param, int),
             (float_param, float),
         ]
+        pyanalyze_suggestions = {}
+        if pyanalyze_report is not None:
+            with open(pyanalyze_report) as f:
+                data = json.load(f)
+            for failure in data:
+                if "lineno" not in failure or "col_offset" not in failure:
+                    continue
+                metadata = failure.get("extra_metadata")
+                if not metadata:
+                    continue
+                if "suggested_type" not in metadata or "imports" not in metadata:
+                    continue
+                if failure.get("code") not in (
+                    "suggested_parameter_type",
+                    "suggested_return_type",
+                ):
+                    continue
+                pyanalyze_suggestions[
+                    (
+                        failure["absolute_filename"],
+                        failure["lineno"],
+                        failure["col_offset"],
+                    )
+                ] = metadata
         self.state = State(
             annotate_optionals=[NamedParam.make(s) for s in annotate_optional]
             if annotate_optional
@@ -188,8 +294,14 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
             param_types={typ for param, typ in param_type_pairs if param},
             annotate_magics=annotate_magics,
             annotate_imprecise_magics=annotate_imprecise_magics,
+            pyanalyze_suggestions=pyanalyze_suggestions,
+            only_without_imports=only_without_imports,
         )
 
+    def is_stub(self) -> bool:
+        filename = self.context.filename
+        return filename is not None and filename.endswith(".pyi")
+
     def visit_FunctionDef(self, node: libcst.FunctionDef) -> None:
         self.state.seen_return_statement.append(False)
         self.state.seen_raise_statement.append(False)
@@ -221,7 +333,9 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
     def leave_FunctionDef(
         self, original_node: libcst.FunctionDef, updated_node: libcst.FunctionDef
     ) -> libcst.CSTNode:
-        is_asynq = any(is_asynq_decorator(dec) for dec in original_node.decorators)
+        kinds = {get_decorator_kind(decorator) for decorator in updated_node.decorators}
+        is_asynq = DecoratorKind.asynq in kinds
+        is_abstractmethod = DecoratorKind.abstractmethod in kinds
         seen_return = self.state.seen_return_statement.pop()
         seen_raise = self.state.seen_raise_statement.pop()
         seen_yield = self.state.seen_yield.pop()
@@ -233,6 +347,33 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
         if original_node.returns is not None:
             return updated_node
 
+        if self.state.pyanalyze_suggestions and self.context.filename:
+            # Currently pyanalyze gives the lineno of the first decorator
+            # and libcst that of the def.
+            # TODO I think the AST behavior changed in later Python versions.
+            if original_node.decorators:
+                lineno_node = original_node.decorators[0]
+            else:
+                lineno_node = original_node
+            pos = self.get_metadata(
+                PositionProvider, lineno_node, _DEFAULT_CODE_RANGE
+            ).start
+            key = (self.context.filename, pos.line, pos.column)
+            suggestion = self.state.pyanalyze_suggestions.get(key)
+            if suggestion is not None and not (
+                suggestion["imports"] and self.state.only_without_imports
+            ):
+                for import_line in suggestion["imports"]:
+                    if "." not in import_line:
+                        AddImportsVisitor.add_needed_import(self.context, import_line)
+                    else:
+                        mod, name = import_line.rsplit(".", maxsplit=1)
+                        AddImportsVisitor.add_needed_import(self.context, mod, name)
+                annotation = libcst.Annotation(
+                    annotation=libcst.parse_expression(suggestion["suggested_type"])
+                )
+                return updated_node.with_changes(returns=annotation)
+
         if self.state.annotate_magics:
             if name in SIMPLE_MAGICS:
                 return updated_node.with_changes(
@@ -255,6 +396,8 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
             and not seen_raise
             and not seen_return
             and (is_asynq or not seen_yield)
+            and not is_abstractmethod
+            and not self.is_stub()
         ):
             return updated_node.with_changes(
                 returns=libcst.Annotation(annotation=libcst.Name(value="None"))
@@ -363,6 +506,26 @@ class AutotypeCommand(VisitorBasedCodemodCommand):
             return updated_node
         if original_node.annotation is not None:
             return updated_node
+        if self.state.pyanalyze_suggestions and self.context.filename:
+            pos = self.get_metadata(
+                PositionProvider, original_node, _DEFAULT_CODE_RANGE
+            ).start
+            key = (self.context.filename, pos.line, pos.column)
+            suggestion = self.state.pyanalyze_suggestions.get(key)
+            if suggestion is not None and not (
+                suggestion["imports"] and self.state.only_without_imports
+            ):
+                for import_line in suggestion["imports"]:
+                    if "." in import_line:
+                        AddImportsVisitor.add_needed_import(self.context, import_line)
+                    else:
+                        mod, name = import_line.rsplit(".", maxsplit=1)
+                        AddImportsVisitor.add_needed_import(self.context, mod, name)
+                annotation = libcst.Annotation(
+                    annotation=libcst.parse_expression(suggestion["suggested_type"])
+                )
+                return updated_node.with_changes(annotation=annotation)
+
         if original_node.default is not None:
             default_type = type_of_expression(original_node.default)
             if default_type is not None and default_type in self.state.param_types:
@@ -437,6 +600,29 @@ def type_of_expression(expr: libcst.BaseExpression) -> Optional[Type[object]]:
             return None
     elif isinstance(expr, libcst.Name) and expr.value in ("True", "False"):
         return bool
+    elif isinstance(expr, libcst.UnaryOperation) and isinstance(
+        expr.operator, libcst.Not
+    ):
+        return bool
+    elif isinstance(expr, libcst.BinaryOperation):
+        left = type_of_expression(expr.left)
+        if left in (str, bytes) and isinstance(expr.operator, libcst.Modulo):
+            return left
+        return None
+    elif isinstance(expr, libcst.BooleanOperation):
+        left = type_of_expression(expr.left)
+        right = type_of_expression(expr.right)
+        # For AND and OR, if both types are the same, we can infer that type.
+        if left == right:
+            return left
+        else:
+            return None
+    elif isinstance(expr, libcst.Comparison):
+        types = {type(comp.operator) for comp in expr.comparisons}
+        # Only these are actually guaranteed to return bool
+        if types <= {libcst.In, libcst.Is, libcst.IsNot, libcst.NotIn}:
+            return bool
+        return None
     elif (
         isinstance(expr, libcst.Call)
         and isinstance(expr.func, libcst.Attribute)
@@ -448,16 +634,26 @@ def type_of_expression(expr: libcst.BaseExpression) -> Optional[Type[object]]:
         return None
 
 
-def is_asynq_decorator(dec: libcst.Decorator) -> bool:
-    """Is this @asynq()?"""
-    if not isinstance(dec.decorator, libcst.Call):
-        return False
-    call = dec.decorator
-    if not isinstance(call.func, libcst.Name):
-        return False
-    if call.func.value != "asynq":
-        return False
-    if call.args:
-        # @asynq() with custom arguments may do something unexpected
-        return False
-    return True
+def get_decorator_kind(dec: libcst.Decorator) -> Optional[DecoratorKind]:
+    """Is this @asynq() or @abstractmethod?"""
+    if isinstance(dec.decorator, libcst.Call):
+        call = dec.decorator
+        if not isinstance(call.func, libcst.Name):
+            return None
+        if call.func.value != "asynq":
+            return None
+        if call.args:
+            # @asynq() with custom arguments may do something unexpected
+            return None
+        return DecoratorKind.asynq
+    elif isinstance(dec.decorator, libcst.Name):
+        if dec.decorator.value == "abstractmethod":
+            return DecoratorKind.abstractmethod
+    elif isinstance(dec.decorator, libcst.Attribute):
+        if (
+            dec.decorator.attr.value == "abstractmethod"
+            and isinstance(dec.decorator.value, libcst.Name)
+            and dec.decorator.value.value == "abc"
+        ):
+            return DecoratorKind.abstractmethod
+    return None
